import click
import os
from vinery.io import echo
from vinery.tf import select_workspace


TF_VARS = {
    'project': ("The name of the project.", "vine"),
}
OPTIONS_TF_VARS = {}


for name, (description, default) in TF_VARS.items():
    def _option_tf_var(function: callable):
        def callback(ctx, param, value):
            os.environ[f"TF_VAR_{name.upper()}"] = value
            return value
        
        return click.option(
            f'--{name}', f'-{name[0]}', f'-{name}',
            help=description,
            callback=callback,
            default=default,
            envvar=f"TF_VAR_{name.upper()}",
            required=True,
            show_default=True,
        )(function)
    
    OPTIONS_TF_VARS[name] = _option_tf_var


def option_workspace(function: callable):
    def callback(ctx, param, value):
        if len(value) > 7:
            raise ValueError("Workspace name must be at most 7 characters long.")
        ctx.ensure_object(dict)
        select_workspace(value, ctx.obj['runner'])
        return value

    return click.option(
        '--workspace', '-w', '-workspace',
        default='default',
        callback=callback,
        envvar="TF_VAR_workspace",
        help="The current workspace against which all plans are evaluated/executed.",
        required=True,
        show_default=True,
    )(function)


def options_tf_vars(function: callable):
    for tf_var, option in OPTIONS_TF_VARS.items():
        function = option(function)

    function = option_workspace(function)
    return function



from collections import deque
import matplotlib.pyplot as plt
import networkx as nx
import os
from typing import Self
from vinery.io import DIRECTORIES, echo


class DependencyGraph(nx.DiGraph):
    def from_library(self, path_to_library: str) -> Self:
        """
        Scans all subfolders in the path_to_library directory and builds a dependency graph.
        """
        for (root, _, files) in os.walk(path_to_library, topdown=True):
            if "_deps.conf" in files:
                plan_name = root.split(path_to_library)[-1].strip("/")
                self.add_node(plan_name)

                with open(f"{root}/_deps.conf", "r") as f:
                    dependencies = [line.strip().strip("/") for line in f if (
                        line.strip().strip("/")
                        and not line.startswith("#")  # Ignore comments
                    )]
                    for dependency in dependencies:
                        self.add_edge(dependency, plan_name)
        
        return self
    

    def from_node(self, node: str) -> Self:
        """
        Function to create a new DependencyGraph object from a single node.
        """
        self.add_node(node)
        return self


    def from_nodes(self, nodes: set[str]) -> Self:
        """
        Function to create a new DependencyGraph object from a single node.
        """
        self.add_nodes_from(nodes)
        return self
    
    
    def wsubgraph(self, nodes: set[str]) -> Self:
        """
        'w' stands for writable.
        Whereas subgraph() creates a Graph view,
        wsubgraph() creates a completely new DependencyGraph object.
        """
        wsubgraph = DependencyGraph().from_nodes(nodes)

        for u, v in self.edges:
            if u in nodes and v in nodes:
                wsubgraph.add_edge(u, v)

        return wsubgraph
    

    def find_all_dependencies(self, node: str, visited=None) -> set:
        """
        Utility function to find all dependencies (parent nodes) recursively
        """

        if visited is None:
            visited = set()

        # Add the current node to the visited set
        visited.add(node)

        # Recursively find all dependencies (parents) of the current node
        for parent in self.predecessors(node):
            if parent not in visited:
                visited = self.find_all_dependencies(parent, visited)

        return visited
    

    def from_nodes_wsubgraph(self, nodes: tuple[str]) -> Self:
        """
        Function to create a new DependencyGraph object from a subgraph,
        given any number of nodes - which the wsubgraph must include.
        """
        # Get all dependencies (parents) for each node
        from itertools import chain
        dependencies = set(
            chain.from_iterable(
                self.find_all_dependencies(node) for node in nodes
            )
        )

        return self.wsubgraph(dependencies)


    def save_to_png(self, target_directory: str = DIRECTORIES["output"]) -> None:
        plt.figure(figsize=(8, 5))
        nx.draw(self, with_labels=True, node_color="lightblue", edge_color="gray", arrowsize=20)
        plt.savefig(os.path.join(target_directory, "graph.png"), dpi=300)

        echo(f"graph.png was saved to {target_directory}.", log_level="INFO")


    def subtract(self, other: Self) -> Self:
        """
        Subtracts nodes and edges from the current graph.
        """
        self.remove_nodes_from(other.nodes)
        self.remove_edges_from(other.edges)
        
        return self
    

    def add(self, other: Self) -> Self:
        """
        Adds nodes and edges from another graph.
        """
        self.add_nodes_from(other.nodes)
        self.add_edges_from(other.edges)
        
        return self
    

    def sorted_list(self, reverse: bool = False) -> list[str]:
        """
        Returns a list of nodes, sorted using Kahn's algorithm.
        By default, the list is sorted from the root to the leaves, by level.
        If reverse=True, the list is sorted from the leaves to the root.
        """
        # Compute in/out-degree (number of incoming/outgoing edges for each node)
        # to start processing root nodes (no incoming) or leaf nodes (no outgoing)
        # depending on traversal direction
        degree = {
            node: (
                self.in_degree(node)
                if not reverse
                else self.out_degree(node)
            ) for node in self.nodes
        }

        # Start with root nodes (in-degree 0) if forward, or leaf nodes (out-degree 0) if reversed
        queue = deque([node for node in self.nodes if degree[node] == 0])

        sorted_nodes = []
        while queue:
            node = queue.popleft()  # Process the next available node
            sorted_nodes.append(node)

            # Get the node's relative
            relatives = self.predecessors(node) if reverse else self.successors(node)

            for relative in relatives:
                degree[relative] -= 1  # Reduce the in-degree (or out-degree for reverse)
                
                # If the node's relative now has no remaining dependencies,
                # add the relative to the queue
                if degree[relative] == 0:
                    queue.append(relative)

        return sorted_nodes



from vinery.io import DIRECTORIES, setup_directories, set_log_level, setup_library, echo
import os


def setup(log_level: str, path_to_library: str):
    """
    Set up the vinery CLI.
    """
    if all([os.path.isdir(dir) for dir in DIRECTORIES.values()]):
        echo(f"Working directories already exist.", log_level="DEBUG")
    else:
        setup_directories()
        echo("Working directories created successfully.", log_level="INFO")

    set_log_level(log_level)

    if os.path.isdir(f"{path_to_library}/default"):
        echo(f"Library already exists at {path_to_library}.", log_level="DEBUG")
    else:
        setup_library(path_to_library)
        echo(f"Library plans copied to {path_to_library}.", log_level="INFO")

    echo(f"--log-level: {log_level}", log_level="DEBUG")
    echo(f"--path-to-library: {path_to_library}", log_level="DEBUG")



import click
from typing import Literal
from vinery.cli.arguments import argument_plan
from vinery.io import echo


def option_auto_approve(function: callable):
    return click.option(
        '--auto-approve', '-a', '-auto-approve',
        default=False,
        is_flag=True,
        help="Pass -auto-approve flag to 'RUNNER apply/destroy'."
    )(function)


def option_path_to_library(function: callable):    
    return click.option(
        '--path-to-library', '-p', '-path-to-library',
        help='Path to the directory with all infrastructure plans.',
        default='./library',
        envvar='VINERY_PATH_TO_LIBRARY',
        required=True,
        show_default=True,
    )(function)


def option_runner(function: callable):
    """
    Automatically checks if the runner is installed.
    """
    from vinery.tf import load_runners

    def callback(ctx, param, value: Literal["terraform", "tofu"]):
        runners_available = load_runners()
        if value not in runners_available:
            echo(f"Runner '{value}' is not installed.", log_level="ERROR")
        
        ctx.ensure_object(dict)
        ctx.obj['runner'] = value
        return value

    return click.option(
        '--runner', '-r', '-runner',
        help='Select the preferred runner for managing infrastructure.',
        default='terraform',
        show_default=True,
        callback=callback
    )(function)


def option_recursive(function: callable):
    return click.option(
        '--recursive', '-rr', '-recursive',
        help='Apply the command recursively.',
        default=True,
        show_default=True,
        required=True
    )(function)


def option_upgrade(function: callable):
    return click.option(
        '--upgrade', '-u', '-upgrade',
        default=False,
        is_flag=True,
        help="Pass -upgrade flag to 'RUNNER init'."
    )(function)


def options_tf(function: callable):
    function = option_upgrade(function)
    function = option_recursive(function)
    function = option_runner(function)
    function = argument_plan(function)

    return function



import os
import pytest
from vinery.dependency_graph import DependencyGraph
import networkx as nx


@pytest.fixture
def graph(tmp_path):
    # Create a fake folder structure with dependency config files
    (tmp_path / "A").mkdir()
    (tmp_path / "B").mkdir()
    (tmp_path / "C").mkdir()
    (tmp_path / "C/D").mkdir()  # Nested folder

    # Create dependency files
    (tmp_path / "A" / "_deps.conf").write_text("")        # A has no dependencies
    (tmp_path / "B" / "_deps.conf").write_text("#commentlikeandsubscribe\nA\n")     # B depends on A
    (tmp_path / "C" / "_deps.conf").write_text("/A\nB/\n//")  # C depends on A and B
    (tmp_path / "C/D" / "_deps.conf").write_text("A\nC\n")  # D depends on A and C (and, indirectly, B)

    # Return the DependencyGraph instance
    return DependencyGraph().from_library(str(tmp_path))


def test_dependency_graph_ignores_line_comments(graph):
    # Assertions
    assert isinstance(graph, DependencyGraph)
    assert "#commentlikeandsubscribe" not in set(graph.nodes)


def test_dependency_graph_ignores_trailing_and_leading_slashes(graph):
    # Assertions
    assert isinstance(graph, DependencyGraph)
    assert "/A" not in set(graph.nodes)
    assert "B/" not in set(graph.nodes)
    assert "//" not in set(graph.nodes)
    assert "A" in set(graph.nodes)
    assert "B" in set(graph.nodes)


def test_dependency_graph_mirrors_the_file_structure(graph, tmp_path):
    """
    Test that build_dependency_graph correctly constructs the dependency graph.
    """
    # Assertions
    assert isinstance(graph, DependencyGraph)
    assert set(graph.nodes) == {"A", "B", "C", "C/D"}  # All nodes exist
    # Also, a node's name consists of the relative path from the tmp_path
    assert set(graph.edges) == {
        ("A", "B"),
        ("A", "C"),
        ("A", "C/D"),
        ("B", "C"),
        ("C", "C/D")
    }  # Correct dependencies


def test_save_graph_png_file_exists(graph, tmp_path):
    graph.save_to_png(target_directory=str(tmp_path))
    assert os.path.isfile(os.path.join(str(tmp_path), "graph.png"))


def test_find_all_dependencies(graph):
    # Test when starting from 'A'
    dependencies_a = graph.find_all_dependencies('A')
    assert dependencies_a == {'A'}  # A has no dependencies, just itself

    # Test when starting from 'B'
    dependencies_b = graph.find_all_dependencies('B')
    assert dependencies_b == {'A', 'B'}  # B depends on A only

    # Test when starting from 'C'
    dependencies_c = graph.find_all_dependencies('C')
    assert dependencies_c == {'A', 'B', 'C'}  # C depends on A and B

    # Test when starting from 'C/D'
    dependencies_d = graph.find_all_dependencies('C/D')
    assert dependencies_d == {'A', 'B', 'C', 'C/D'}  # D depends on all nodes, even B (indirectly)


def test_wsubgraph(graph):
    """Test that wsubgraph creates a correct writable subgraph."""
    subset_of_nodes = {"A", "B", "C"}  

    # Generate the writable subgraph
    subgraph = graph.wsubgraph(subset_of_nodes)

    # Check that only the selected nodes are present
    assert set(subgraph.nodes) == subset_of_nodes, \
        f"Expected nodes {subset_of_nodes}, but got {set(subgraph.nodes)}"

    # Check that edges are correctly retained (only within selected nodes)
    expected_edges = [(u, v) for u, v in graph.edges if u in subset_of_nodes and v in subset_of_nodes]
    assert set(subgraph.edges) == set(expected_edges), \
        f"Expected edges {expected_edges}, but got {set(subgraph.edges)}"

    # Ensure that the returned subgraph is a completely new instance, not a view
    assert subgraph is not graph, "wsubgraph() should return a new instance, not a view"

    # Ensure modifications to the subgraph do not affect the original graph
    subgraph.add_node("X")
    assert "X" not in graph.nodes, "Modifying the subgraph should not affect the original graph"


def test_from_nodes_wsubgraph(graph):
    """
    Test the from_nodes_wsubgraph function.
    Works with
    """

    # Test for 'A', which should only return node A (no dependencies)
    new_graph_a = graph.from_nodes_wsubgraph(['A'])
    assert sorted(new_graph_a.nodes) == ['A']
    assert list(new_graph_a.edges) == []  # No edges in the subgraph for A
    
    # Test for 'B', which should return a subgraph with nodes A, and B
    new_graph_b = graph.from_nodes_wsubgraph({'B'})
    assert sorted(new_graph_b.nodes) == sorted(['A', 'B'])
    assert ('A', 'B') in new_graph_b.edges  # Edge from A to B should be present

    # Test for ('C', 'A'), which should return a subgraph with nodes A, B, and C
    new_graph_c = graph.from_nodes_wsubgraph(('C', 'A'))
    assert sorted(new_graph_c.nodes) == sorted(['A', 'B', 'C'])
    assert ('A', 'B') in new_graph_c.edges  # Edge from A to C should be present
    assert ('A', 'C') in new_graph_c.edges  # Edge from A to C should be present
    assert ('B', 'C') in new_graph_c.edges  # Edge from A to C should be present

    # Test for 'D', which should return all nodes
    new_graph_d = graph.from_nodes_wsubgraph(['C/D'])
    assert sorted(graph.nodes) == sorted(new_graph_d.nodes)
    assert sorted(graph.edges) == sorted(new_graph_d.edges)


def test_sorting(graph):
    """Test normal topological sorting (root to leaves)."""
    result = graph.sorted_list()
    assert all(result.index(u) < result.index(v) for u, v in graph.edges), \
        f"Result {result} is not a valid topological order."


def test_sorting_reversed(graph):
    """Test reverse topological sorting (leaves to root)."""
    result = graph.sorted_list(reverse=True)
    assert all(result.index(u) > result.index(v) for u, v in graph.edges), \
        f"Result {result} is not a valid topological order"



import click
from datetime import datetime
from importlib.resources import files
from pathlib import Path
import os
import shutil


DIRECTORIES = {
    'tmp': '/tmp/vinery',
    'output': f'{Path().resolve()}/outputs'
}
LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "SUCCESS", "ERROR"]


def setup_directories() -> None:
    for dir in DIRECTORIES.values():
        Path(dir).mkdir(parents=True, exist_ok=True)


def set_log_level(log_level: str) -> None:
    if log_level not in LOG_LEVELS:
        raise ValueError(f"Invalid log level: {log_level}. Accepted values: {LOG_LEVELS}")
    os.environ["VINE_LOG_LEVEL"] = os.getenv("VINE_LOG_LEVEL", log_level)


def setup_library(path_to_library: str) -> None:
    package_library_path = files("vinery").joinpath("library")
    # Copy files from package to user-specified directory
    shutil.copytree(package_library_path, os.path.join(path_to_library), dirs_exist_ok=True)


def read_file(filename: str, dir: str = 'tmp') -> set[str]:
    try:
        with open(os.path.join(DIRECTORIES[dir], filename), "r") as f:
            return set(f.readlines())
    except FileNotFoundError:
        echo(f"File {filename} not found in {DIRECTORIES[dir]}.", log_level="WARNING")
        return set()


def update_file(filename: str, new_lines: list[str], dir: str = 'tmp') -> set[str]:
    contents = read_file(filename, dir)

    for line in new_lines:
        contents.add(line)
    
    with open(os.path.join(DIRECTORIES[dir], filename), "w") as f:
        f.writelines(contents)


def echo(message: str, log_level: str = "INFO") -> None:
    """
    Custom logging function with color-coded output.
    """
    if LOG_LEVELS.index(log_level) < LOG_LEVELS.index(os.getenv("VINE_LOG_LEVEL", "INFO")):
        return  # Suppress messages below the global log level

    message = f"{datetime.now().time().isoformat(timespec='seconds')} vinery: [{log_level}] {message}"

    match log_level:
        case "DEBUG":
            click.secho(message)
        case "INFO":
            click.secho(message, fg="blue", bold=True)
        case "WARNING":
            click.secho(message, fg="yellow")
        case "SUCCESS":
            click.secho(message, fg="green", bold=True)
        case "ERROR":
            click.secho(message, fg="red", bold=True, err=True)


import os
import subprocess
from vinery.dependency_graph import DependencyGraph
from vinery.io import read_file, update_file, echo

SUPPORTED_RUNNERS=["terraform", "tofu"]


class RunnerNotFoundError(Exception):
    pass


def load_runners() -> list[str]:
    runners = [
        runner for runner in SUPPORTED_RUNNERS
        if subprocess.run(
            args=["which", runner],
            capture_output=True
        ).stdout.decode().strip()
    ]

    if not runners:
        raise RunnerNotFoundError("ERROR: No runner is installed.")
    
    return runners


def list_workspaces(runner: str) -> list[str]:
    output = subprocess.run(
        args=[runner, "workspace", "list"],
        check=True,
        capture_output=True,
    ).stdout.decode().replace("*", "")

    return [line.strip() for line in output.split("\n") if line]


def select_workspace(workspace: str, runner: str) -> int:
    list_of_existing_workspaces = list_workspaces(runner)
    cmd = "new" if workspace not in list_of_existing_workspaces else "select"
    try:
        subprocess.run(
            args=[runner, "workspace", cmd, workspace],
            check=True,
        )
        echo(f"Selected workspace '{workspace}'.", log_level="INFO")
        return 0
    
    except subprocess.CalledProcessError:
        echo(f"Failed to select workspace '{workspace}'.", log_level="ERROR")
        return 1


def tf(
    plan: str,
    runner: str,
    cmd: str,
    path_to_library: str,
    save_output: bool = False,
) -> int:
    cmd = f"{runner} {cmd}"
    echo(f"tf('{plan}', '{cmd}', '{path_to_library}', {save_output})", log_level="DEBUG")
    echo(f"Running command '{cmd}' for plan '{plan}'.", log_level="INFO")
    
    try:
        output = subprocess.run(
            args=cmd,
            cwd=os.path.join(path_to_library, plan),
            check=True,
            capture_output=save_output,
            shell=True,
        )
        if save_output:
            update_file(
                f"{cmd.split(' ')[1]}_{plan.replace('/', '_')}.log",
                [output.stdout.decode()],
                dir='output'
            )
        echo(f"Command '{cmd}' for plan '{plan}' was successful!", log_level="SUCCESS")
        return 0
    
    except subprocess.CalledProcessError:
        echo(f"Command '{cmd}' failed for plan {plan}!", log_level="ERROR")
        return 1


def tf_loop(
    graph_of_plans_to_run: DependencyGraph,
    *args,
    reverse: bool = False,
    **kwargs
) -> DependencyGraph:
    return graph_of_plans_to_run.wsubgraph({
        plan for plan in graph_of_plans_to_run.sorted_list(reverse)
        if tf(plan, *args, **kwargs) == 0
    })


def with_dependency_graph(function):
    """
    Decorator that creates the dependency graph of all relevant plans.
    """
    def wrapper(plan: tuple[str], path_to_library, recursive, *args, **kwargs):
        graph_of_plans = (
            DependencyGraph()
            .from_library(path_to_library)
            .from_nodes_wsubgraph(plan)
        ) if recursive else DependencyGraph().from_node(plan)

        return function(graph_of_plans, path_to_library, *args, **kwargs)
    
    return wrapper


@with_dependency_graph
def init(graph_of_plans, path_to_library, runner, upgrade) -> DependencyGraph:
    graph_of_plans_initialized = graph_of_plans.wsubgraph(
        read_file("init_status") if not upgrade else set()
    )
    graph_of_plans_to_initialize = graph_of_plans.subtract(graph_of_plans_initialized)

    if not graph_of_plans_to_initialize:
        echo("No plans require initialization. Did you mean to run -upgrade?", log_level="INFO")
        return graph_of_plans.wsubgraph(graph_of_plans_initialized.nodes)

    graph_of_plans_initialized.add(tf_loop(
        graph_of_plans_to_initialize,
        runner, f"init{' -upgrade' if upgrade else ''}", path_to_library,
    ))

    update_file("init_status", graph_of_plans_initialized.nodes)

    return graph_of_plans_initialized


def with_tf_init(function):
    """
    Decorator that runs 'init' before the function.
    """
    @with_dependency_graph
    def wrapper(graph_of_plans, path_to_library, runner, upgrade, *args, **kwargs):
        graph_of_plans_initialized = init(graph_of_plans, path_to_library, runner, upgrade=upgrade)
        return function(graph_of_plans_initialized, path_to_library, runner, *args, **kwargs)

    return wrapper


@with_tf_init
def validate(graph_of_plans_initialized, path_to_library, runner, json) -> DependencyGraph:
    return tf_loop(
        graph_of_plans_initialized,
        runner, f"validate{' -json' if json else ''}", path_to_library,
        save_output=json,
    )


@with_tf_init
def plan(graph_of_plans_initialized, path_to_library, runner) -> DependencyGraph:
    return tf_loop(
        graph_of_plans_initialized,
        runner, "plan", path_to_library,
    )


@with_tf_init
def apply(graph_of_plans_initialized, path_to_library, runner, auto_approve) -> DependencyGraph:
    return tf_loop(
        graph_of_plans_initialized,
        runner, f"apply{' -auto-approve' if auto_approve else ''}", path_to_library,
    )


@with_tf_init
def destroy(graph_of_plans_initialized, path_to_library, runner, auto_approve) -> DependencyGraph:
    return tf_loop(
        graph_of_plans_initialized,
        runner, f"destroy{' -auto-approve' if auto_approve else ''}", path_to_library, reverse=True,
    )



import click
from vinery.io import echo


def argument_plan(function: callable):
    help_text = """
    PLAN: apply command to the PLAN dependency tree, including PLAN.
    Can take multiple plans as arguments.
    """

    if function.__doc__:
        function.__doc__ += help_text
    else:
        function.__doc__ = help_text
    
    def callback(ctx, param, value):
        if value is None:
            echo("At least ONE plan is required.", log_level="ERROR")
        return [plan.strip("/") for plan in value]
    
    return click.argument(
        'plan',
        callback=callback,
        nargs=-1  # Variadic argument
    )(function)



import click
import subprocess
from vinery.cli.options import option_path_to_library, options_tf, option_runner, option_auto_approve
from vinery.cli.options_tf_vars import options_tf_vars
from vinery.io import LOG_LEVELS
from vinery.cli.setup import setup
from vinery import tf


@click.group()
@click.option(
    '--log_level', '-l',
    default="INFO",
    envvar="VINE_LOG_LEVEL",
    help=f"""
    Set the global log level for the CLI.
    Can be overridden by setting the VINE_LOG_LEVEL environment variable.
    
    Accepted values: {LOG_LEVELS}
    """,
    show_default=True
)
@option_path_to_library
@click.pass_context
def cli(ctx, log_level: str, path_to_library: str):
    """
    Manage infrastructure plans.
    """
    setup(log_level, path_to_library)

    ctx.ensure_object(dict)
    ctx.obj["path_to_library"] = path_to_library


########################
# fmt
@cli.command(
    help="Recursively format all infrastructure plans."
)
@option_runner
@click.pass_context
def fmt(ctx, runner: str):
    subprocess.run(args=[runner, "fmt", "-recursive"], cwd=ctx.obj["path_to_library"])


########################
# init
@cli.command()
@options_tf
@click.pass_context
def init(ctx, plan: str, runner: str, recursive: bool, upgrade: bool):
    """
    Initialize all infrastructure plans.
    """
    tf.init(plan, ctx.obj["path_to_library"], recursive, runner, upgrade)


########################
# validate
@cli.command()
@options_tf
@click.option(
    '--json', '-j', '-json',
    default=False,
    is_flag=True,
    help="""
    Pass -json flag to 'RUNNER plan'.
    Additionally, saves JSON output to a file.
    """
)
@click.pass_context
def validate(ctx, plan: str, runner: str, recursive: bool, upgrade: bool, json: bool):
    """
    Validate plans' syntax and correctness.
    By default, runs 'RUNNER init -upgrade' prior to execution.
    Plans that fail to 'init' are not validated.
    """
    tf.validate(plan, ctx.obj["path_to_library"], recursive, runner, upgrade, json)


########################
# plan
@cli.command()
@options_tf
@options_tf_vars
@click.pass_context
def plan(ctx, plan: str, runner: str, recursive: bool, upgrade: bool):
    """
    Execute a dry run of all infrastructure plans,
    showing what changes would be made.
    """
    tf.plan(plan, ctx.obj["path_to_library"], recursive, runner, upgrade)


########################
# apply
@cli.command()
@options_tf
@option_auto_approve
@options_tf_vars
@click.pass_context
def apply(ctx, plan: str, runner: str, recursive: bool, upgrade: bool, auto_approve: bool):
    """
    Apply the plans, building the infrastructure and applying any latent changes.
    """
    tf.apply(plan, ctx.obj["path_to_library"], recursive, runner, upgrade, auto_approve)


########################
# destroy
@cli.command()
@options_tf
@option_auto_approve
@click.pass_context
def destroy(ctx, plan: str, runner: str, recursive: bool, upgrade: bool, auto_approve: bool):
    """
    Destroy all infrastructure described in the associated set of plans.
    """
    tf.destroy(plan, ctx.obj["path_to_library"], recursive, runner, upgrade, auto_approve)



import pytest
import subprocess
from unittest.mock import patch, MagicMock
from vinery.tf import load_runners, SUPPORTED_RUNNERS, tf


def test_load_runners_returns_all_runners_if_all_runners_exist():
    # All supported runners must exist in the test environment!
    assert load_runners() == SUPPORTED_RUNNERS


@pytest.fixture
def mock_subprocess_run():
    """Mock subprocess.run for controlled command execution."""
    with patch("subprocess.run") as mock_run:
        yield mock_run


@pytest.fixture
def mock_update_file():
    """Mock update_file function to prevent actual file writing."""
    with patch("vinery.tf.update_file") as mock_update:
        yield mock_update


@pytest.fixture
def mock_echo():
    """Mock echo function to suppress output in tests."""
    with patch("vinery.tf.echo") as mock_echo_fn:
        yield mock_echo_fn


def test_tf_success(mock_subprocess_run, mock_update_file, mock_echo):
    """Test tf() when the command runs successfully."""
    mock_subprocess_run.return_value = MagicMock(returncode=0, stdout=b"Success output")

    result = tf("my_plan", "my_runner", "my_cmd", "my_path")

    assert result == 0, "Expected return code 0 for success"
    mock_subprocess_run.assert_called_once()
    mock_echo.assert_any_call("Command 'my_runner my_cmd' for plan 'my_plan' was successful!", log_level="SUCCESS")


def test_tf_failure(mock_subprocess_run, mock_update_file, mock_echo):
    """Test tf() when the command fails."""
    mock_subprocess_run.side_effect = subprocess.CalledProcessError(1, "terraform apply")

    result = tf("my_plan", "my_runner", "my_cmd", "my_path")

    assert result == 1, "Expected return code 1 for failure"
    mock_subprocess_run.assert_called_once()
    mock_echo.assert_any_call("Command 'my_runner my_cmd' failed for plan my_plan!", log_level="ERROR")


def test_tf_save_output(mock_subprocess_run, mock_update_file, mock_echo):
    """#Test tf() when save_output=True."""
    mock_subprocess_run.return_value = MagicMock(returncode=0, stdout=b"Saved output")

    result = tf("my_plan", "my_runner", "my_cmd", "my_path", save_output=True)

    assert result == 0, "Expected return code 0 when saving output"
    mock_update_file.assert_called_once_with(
        "my_cmd_my_plan.log", ["Saved output"], dir="output"
    )
    mock_echo.assert_any_call("Command 'my_runner my_cmd' for plan 'my_plan' was successful!", log_level="SUCCESS")
